#!/bin/bash 

# ====================================================
#
# DVD Builder script by Dik Takken
# 
# License: GPL
#
# ====================================================

# TODO:
#
# * Check if files + menus will fit on DVD disk
#
# * Clean / Restructure code
#
# * Remove MPlayer dependency, ffmpeg can do better 
#   video feature detection.
#
# * Add second total-progress bar in GUI
#
# * Split GUI and script in authoring and menu creation
#
# * Support for timecode-based chapter definitions. Also
#   Nice for DV Assembler users.


# Set this variable to true for debug/test mode.
# This will generate very short menu animations and
# will not remove any temporary files.

# DON'T forget to remove temporary files when you turn
# debugging off while using the Kommander GUI.

Debug="false"


# ----------------------------------------------------------
# ROUTINES TO COMMUNICATE ERRORS/WARNINGS TO THE USER
# ----------------------------------------------------------

# Present an error message to the user and abort.
# Argument $1 is error message.

PrintError ()
{
 X=`type -p kdialog`

 if test $X && test -x $X && [ -n "$DISPLAY" ]; then

  kdialog --error "$1"

  UpdateTask "Aborted."
  UpdateProgress 0

 else

  echo
  echo "---- RECEIVED ERROR MESSAGE ----"
  echo
  echo -e "$1"
  echo
  echo "--------------------------------"

 fi

 exit 1
}

# Present a warning message to the user
# Argument $1 is warning message.

PrintWarning ()
{
 X=`type -p kdialog`

 if test $X && test -x $X && [ -n "$DISPLAY" ]; then
  kdialog --warningcontinuecancel "$1"
  if [ ! "$?" -eq 0 ]; then
  echo
  echo "USER ABORTED PROGRAM"
  echo
  exit 1
  fi
  
 else
  echo
  echo "---- WARNING ----"
  echo
  echo -e "$1"
  echo
  echo "-----------------"
 fi
}

# ----------------------------------------------------------
# CHECKS TO SEE IF WE HAVE ALL THE REQUIRED TOOLS AVAILABLE.
# ----------------------------------------------------------

Requirements=( convert composite mogrify identify sox mplex dvdauthor png2yuv mplayer tcprobe mpeg2enc mp2enc bc kdialog konsole )

PrettyRequirements=( "ImageMagick >= 6.1" "MJPEGTools" "Sox" "DVDAuthor >= 0.6" "MPlayer >= 1.0pre4" "Transcode" "bc" )

# Check if tool $1 is available.

ToolCheck ()
{
 echo -ne "Checking for \"$1\" ...\t"
	X=`type -p $1`
	if test $X && test -x $X;
		then echo " found." ; return 0
		else echo " MISSING" ; return 1
	fi
}

# Print an error when a required utility is missing

Missing ()
{
 PrintError "The utility \"$1\" has not been installed on your computer! \n\nThe KDE DVD Authoring Wizard requires the following software packages to be installed:\n\n`
 
 Requirement=0
 while [ -n "${PrettyRequirements[$Requirement]}" ]
 do
  echo -n "* ${PrettyRequirements[$Requirement]}\n"
  Requirement=$[$Requirement + 1]
 done

 `" 

 ChecksOK="false"

}

RunChecks ()
{
 ChecksOK="true"
 Requirement=0
 while [ -n "${Requirements[$Requirement]}" ]
 do
   ToolCheck "${Requirements[$Requirement]}" || Missing "${Requirements[$Requirement]}"
   Requirement=$[$Requirement + 1]
 done
}

# ----------------------------------------------------------
# COMMUNICATION ROUTINES FOR USE WITH KOMMANDER GUI
# ----------------------------------------------------------

# Set the currently executing task to $1.

UpdateTask ()
{
 if [ -n "$DialogDCOPTaskCommand" ]; then
  $DialogDCOPTaskCommand "$1"
 fi
}

# Set the currently executing task to $1.

UpdateProgress ()
{
 if [ -n "$DialogDCOPProgressCommand" ]; then
   $DialogDCOPProgressCommand $1
 fi
}

# ----------------------------------------------------------
# DISK SPACE CHECKING ROUTINES
# ----------------------------------------------------------

# Warn if disk space in directory $1 lower than $2 MB

DiskSpaceWarning ()
{
  FreeSpace=`df  -m "$1" -P | awk '{print $4}' | tail -n 1`

  if [ $FreeSpace -lt $2 ]; then

    PrintWarning "You are low on disk space in this folder:\n\n$1\n\nYou are strongly encouraged to free some disk space before continuing."

    if [ "$?" -eq 1 ]; then
      exit 1
    fi

  fi
}

# Abort if disk space ran out in directory $1

DiskSpaceError ()
{
  FreeSpace=`df  -m "$1" -P | awk '{print $4}' | tail -n 1`

  if [ $FreeSpace -le 1 ]; then
    PrintError "You ran out of disk space in this folder:\n\n$1\n\nPlease free more disk space and try again."
    exit 1
  fi
}

# ----------------------------------------------------------
# SCALE BACKGROUND IMAGE TO FIT SCREEN
# ----------------------------------------------------------

ProcessBackgroundImage ()
{
 if [ "$BuildMenu" == "true" ] && [ $BackgroundType == "Picture" ]; then

 YRes=`identify "$BackgroundPicture" | grep -E -o [0-9]{1,\}x[0-9]{1,\} | head -n 1 | cut -d "x" -f 2`
 XRes=`identify "$BackgroundPicture" | grep -E -o [0-9]{1,\}x[0-9]{1,\} | head -n 1 | cut -d "x" -f 1`

 UpdateTask "Fitting background picture to DVD image size..."
 echo "XRes is $XRes and YRes is $YRes"

 if [ $XRes -gt $YRes ]; then

    echo "Scaling..."
    convert "$BackgroundPicture" -resize $MenuWidth "$WorkingDir/backgroundpicture.png"
    YRes=`identify "$WorkingDir/backgroundpicture.png" | grep -E -o [0-9]{1,\}x[0-9]{1,\} | head -n 1 | cut -d "x" -f 2`

    if [ $YRes -gt $MenuHeight ]; then
      echo "Cropping..."
      convert "$WorkingDir/backgroundpicture.png" -crop ${MenuWidth}x${MenuHeight}+0+$[$[$YRes - $MenuHeight] / 2] "$WorkingDir/backgroundpicture.png"
    fi

    if [ $YRes -lt $MenuHeight ]; then
      echo "Expanding..."
      convert \( -size ${MenuWidth}x${MenuHeight} xc:black "$WorkingDir/backgroundpicture.png" -geometry +0+$[$[$MenuHeight - $YRes] / 2] -compose over -composite \) "$WorkingDir/backgroundpicture.png"
    fi

 else

    echo "Scaling..."
    convert "$BackgroundPicture" -resize x${MenuHeight} "$WorkingDir/backgroundpicture.png"
    XRes=`identify "$WorkingDir/backgroundpicture.png" | grep -E -o [0-9]{1,\}x[0-9]{1,\} | head -n 1 | cut -d "x" -f 1`

    if [ $XRes -gt $MenuWidth ]; then
      echo "Cropping..."
      convert "$WorkingDir/backgroundpicture.png" -crop ${MenuWidth}x${MenuHeight}+$[$[$XRes - $MenuWidth] / 2]+0 "$WorkingDir/backgroundpicture.png"
    fi

    if [ $XRes -lt $MenuWidth ]; then
      echo "Expanding..."
      convert \( -size ${MenuWidth}x${MenuHeight} xc:black "$WorkingDir/backgroundpicture.png" -geometry +$[$[$MenuWidth - $XRes] / 2]+0 -compose over -composite \) "$WorkingDir/backgroundpicture.png"
    fi

  fi

  UpdateTask ""

 fi
}

# ----------------------------------------------------------
# FUNCTION THAT SPLITS SCENES INTO THUMBNAILED FRAMES.
# ----------------------------------------------------------

GenerateThumbFrames ()
{
 echo
 echo "--------------------------------"
 echo "Generating Thumbnail Frames"
 echo "--------------------------------"
 echo

 if [ -d "$WorkingDir/Thumb1" ]; then
   echo "Thumb frames already generated, skipping."

   ThumbWithMostFrames=1
   HighestFrameCount=0

   for ((Thumb=1; Thumb <= NumScenes ; Thumb++))
     do
     PreviousDir="`pwd`"
     cd "$WorkingDir/Thumb$Thumb"

     # Remember how many frames are actually generated. Might be less than requested.
     AvailableFrames[$Thumb]=$[`ls -l | wc -l` - 1 ]

     if [ ${AvailableFrames[$Thumb]} -gt $HighestFrameCount ]; then
       HighestFrameCount=${AvailableFrames[$Thumb]}
       ThumbWithMostFrames=$Thumb
     fi

     echo "Available frames in Thumb $Thumb: ${AvailableFrames[$Thumb]}"
     cd "$PreviousDir"
   done

   echo "Thumb $ThumbWithMostFrames has most frames."
   ThumbFrameNames=( `ls "$WorkingDir/Thumb$ThumbWithMostFrames/"` )
   ForceMenuUpdate="false"
 else

   if [ "$BuildMenu" == "true" ]; then

     echo -n "Generating frames for Thumb "
     ForceMenuUpdate="true"

     ThumbWithMostFrames=1
     HighestFrameCount=0

     for ((Thumb=1; Thumb <= NumScenes ; Thumb++))
       do
       echo -n $Thumb,
       PreviousDir="`pwd`"
       mkdir -p "$WorkingDir/Thumb$Thumb"
       cd "$WorkingDir/Thumb$Thumb"
       rm -f *.png
       UpdateTask "Extracting frames from ${SceneTitles[$Thumb]}..."
       DiskSpaceWarning "`pwd`" $[ $NumSceneFrames / 10 ]

       if [ -z "`tcprobe -i "${SceneFiles[$Thumb]}" 2>&1 | grep "MPEG program stream"`" ]; then
         PrintError "File ${SceneFiles[$Thumb]} is not a valid MPEG2 PS file."
       fi

       ThumbAspect="`tcprobe -i "${SceneFiles[$Thumb]}" 2>&1 | grep -E '\<aspect ratio' | grep -o -E [0-9]{1,\}:[0-9]{1,\}`"

       # Output frames as png thumbnails. Use bicubic non-interlaced scaling.

       if [ $WidescreenDVD == "true" ]; then

         if [ "$ThumbAspect" == "16:9" ]; then

	  mplayer "${SceneFiles[$Thumb]}" -sws 2 -frames $NumSceneFrames -nosound -noframedrop -vo png -vf  pp=lb,scale=$ThumbWidth:$ThumbHeight < /dev/null > /dev/null 2>&1

	 else 

	  mplayer "${SceneFiles[$Thumb]}" -sws 2 -frames $NumSceneFrames -nosound -noframedrop -vo png -vf  pp=lb,crop=768:432,scale=$ThumbWidth:$ThumbHeight < /dev/null > /dev/null 2>&1

	 fi
       else

         if [ "$ThumbAspect" == "16:9" ]; then

	  mplayer "${SceneFiles[$Thumb]}" -sws 2 -frames $NumSceneFrames -nosound -noframedrop -vo png -vf pp=lb,expand=1024:768,scale=$ThumbWidth:$ThumbHeight > /dev/null  2>&1 < /dev/null

	 else

	  mplayer "${SceneFiles[$Thumb]}" -sws 2 -frames $NumSceneFrames -nosound -noframedrop -vo png -vf pp=lb,scale=$ThumbWidth:$ThumbHeight  > /dev/null 2>&1 < /dev/null

	 fi

       fi

       if [ $? -gt 0 ] || [ `ls | wc -l` -lt 1 ]; then

         PrintError "Whoops. Your MPlayer version failed while generating frames from this video file:\n\n${SceneFiles[$Thumb]}"

       fi

       # Remember how many frames are actually generated. Might be less than requested.
       AvailableFrames[$Thumb]=$[`ls -l | wc -l` - 1 ]

       if [ ${AvailableFrames[$Thumb]} -gt $HighestFrameCount ]; then
         HighestFrameCount=${AvailableFrames[$Thumb]}
         ThumbWithMostFrames=$Thumb
       fi
       echo "Available frames in clip $Thumb: ${AvailableFrames[$Thumb]}"

       DiskSpaceError "`pwd`"

       cd "$PreviousDir"
       UpdateProgress $[$Thumb * 100 / $NumScenes]
     done

     ThumbFrameNames=( `ls "$WorkingDir/Thumb$ThumbWithMostFrames/"` )

     DecorateThumbFrames

     ThumbFrameNames=( `ls "$WorkingDir/Thumb$ThumbWithMostFrames/"` )

     if [ "$BackgroundType" == "Movie" ]; then
       if [ -d "BackgroundMovie" ]; then
         echo "Background movie frames already generated. Skipping."
       else 

         # Output frames of background movie as png thumbnails.

         UpdateTask "Extracting frames from background movie..."

         PreviousDir="`pwd`"
	 mkdir -p "$WorkingDir/BackgroundMovie"
         cd "$WorkingDir/BackgroundMovie"

	 if [ -n "`tcprobe -i "$BackgroundMovie" 2>&1 | grep "MPEG program stream"`" ]; then

	   echo "Background movie is MPEG2 movie."

           MovieAspect="`tcprobe -i "$BackgroundMovie" 2>&1 | grep -E '\<aspect ratio' | grep -o -E [0-9]{1,\}:[0-9]{1,\}`"

	   if [ $WidescreenDVD == "true" ]; then

             if [ "$MovieAspect" == "16:9" ]; then

	       MPlayerFilters="scale=1024:576,pp=hb/vb/dr"

	     else 

	       # We HAVE to deinterlace because we can't detect if MPEG2 file
	       # is interlaced or not... :(

	       MPlayerFilters="kerndeint,crop=768:432,scale=1024:576,pp=hb/vb/dr"

	     fi

	   else

             if [ "$MovieAspect" == "16:9" ]; then

	       # We HAVE to deinterlace because we can't detect if MPEG2 file
	       # is interlaced or not... :(

	       MPlayerFilters="kerndeint,expand=1024:768,scale=768:576,pp=hb/vb/dr"

	     else

	       MPlayerFilters="scale=768:576,pp=hb/vb/dr"

	     fi

           fi
 
	 else

	   PrintError "Background movie is not a valid MPEG2 PS file."

	 fi

         DiskSpaceWarning "`pwd`" $[ $TotalFrameCount ]

	 mplayer "$BackgroundMovie" -sws 2 -frames $TotalFrameCount -nosound -noframedrop -vo png -vf $MPlayerFilters < /dev/null > /dev/null 2>&1

         DiskSpaceError "`pwd`"

         if [ $? -gt 0 ] || [ `ls | wc -l` -lt 1 ]; then

           PrintError "Whoops. Your MPlayer version failed while generating frames from this video file:\n\n$BackgroundMovie"

         fi

         cd "$PreviousDir"
         UpdateTask ""
         UpdateProgress 0
       fi

     fi

   fi 

 fi
}

# ---------------------------------------------------------------
# ADDS NICE DECORATIVE BORDERS TO THE SCENE FRAMES
# ---------------------------------------------------------------

DecorateThumbFrames ()
{
  echo
  echo "--------------------------------"
  echo "Preprocessing Thumbnail Frames:"
  echo "--------------------------------"
  echo

  # Determine the correct mask for the job
  # Might consider scaling the original on demand

  if [ $ThumbDecoration == "SoftEdge" ]; then

    if [ $WidescreenDVD == "true" ]; then
      ThumbMaskImage="./Pictures/maskwide.png"
    else
      ThumbMaskImage="./Pictures/mask.png"
    fi

  elif [ $ThumbDecoration == "FilmRoll" ]; then

    if [ $WidescreenDVD == "true" ]; then
      ThumbMask="./Pictures/rollwidemask.png"
      ThumbBackground="./Pictures/rollwide.png"
    else
      ThumbMask="./Pictures/rollmask.png"
      ThumbBackground="./Pictures/roll.png"
    fi

  else

    return

  fi

  for ((Thumb=1; Thumb <= NumScenes ; Thumb++))
  do
    UpdateTask "Decorating thumbnail frames of ${SceneTitles[$Thumb]}..."

    FirstDecoFrame=0
    LastDecoFrame=${AvailableFrames[$Thumb]}

    if [ "$StaticMenu" == "true" ]; then

      FirstDecoFrame=$[$ThumbFrameCounter % ${AvailableFrames[$Thumb]}]
      LastDecoFrame=$[$FirstDecoFrame + 1]

    fi

    if [ $ThumbDecoration == "SoftEdge" ]; then
      for ((Frame=$FirstDecoFrame; Frame < $LastDecoFrame ; Frame++))
      do
        if [ -n "${ThumbFrameNames[$Frame]}" ] && [ -e "$WorkingDir/Thumb$Thumb/${ThumbFrameNames[$Frame]}" ]; then

	  # Do a Alpha mask composition with clipmask
          composite -compose Dst_In -gravity center "$ThumbMaskImage"  "$WorkingDir/Thumb$Thumb/${ThumbFrameNames[$Frame]}" -matte "$WorkingDir/Thumb$Thumb/${ThumbFrameNames[$Frame]}"

          if [ $(( $Frame % 10 )) == 0 ]; then
	    UpdateProgress $[$Frame * 100 / $NumSceneFrames]
	  fi

        fi
      done
    elif [ $ThumbDecoration == "FilmRoll" ]; then
      for ((Frame=$FirstDecoFrame; Frame < $LastDecoFrame ; Frame++))
      do
        if [ -n "${ThumbFrameNames[$Frame]}" ] && [ -e "$WorkingDir/Thumb$Thumb/${ThumbFrameNames[$Frame]}" ]; then

          # Do a Alpha mask composition with mask.png
          convert \( "$ThumbBackground" \( "$WorkingDir/Thumb$Thumb/${ThumbFrameNames[$Frame]}" "$ThumbMask" -compose Dst_In -gravity center \) -compose over -gravity center -composite \)  "$WorkingDir/Thumb$Thumb/${ThumbFrameNames[$Frame]}"

          if [ $(( $Frame % 10 )) == 0 ]; then
             UpdateProgress $[$Frame * 100 / $NumSceneFrames]
          fi
        fi
      done
    fi
  UpdateProgress 100    
  done
}

# ---------------------------------------------------------------
# CREATE SOME GRAPHICS THAT WE WILL NEED THROUGHOUT THE PROCESS.
# ---------------------------------------------------------------

Initialise ()
{
 echo
 echo "--------------------------------"
 echo "Initialising:"
 echo "--------------------------------"
 echo

 UpdateTask "Initialising frame generator..."

 # Generate a nice title:

 convert -size ${MenuWidth}x$[${ThumbY[1]} - 30] -pointsize 80 xc:'#000000FF' -gravity center -fill $FontColor -stroke $FontOutlineColor -strokewidth 3 -draw "font \"$TitleFont\" text 0,0 '$Title'" -stroke none -draw "font $TitleFont  text 0,0 \"$Title\"" "$WorkingDir/Title.png"

 # And a subtitle:

 convert -size ${MenuWidth}x$[${ThumbY[1]} - 100] -pointsize 24 xc:'#000000FF' -gravity center -fill $FontColor -stroke $FontOutlineColor -strokewidth 3 -draw "font \"$SubtitleFont\" text 0,0 '$Subtitle'" -stroke none -draw "font $SubtitleFont  text 0,0 \"$Subtitle\"" "$WorkingDir/Subtitle.png"

 # Generate labels for all clips:

 echo "Generating Thumbnail Labels"
 for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
   do
     convert -size $[$ThumbWidth + $InterColSpace + $InterColSpace]x37 -pointsize 18 xc:'#000000FF' -gravity center -fill $FontColor -stroke $FontOutlineColor -strokewidth 3 -draw "font \"$LabelFont\" text 0,0 '${SceneTitles[$Thumb]}'" -stroke none -draw "font $LabelFont text 0,0 \"${SceneTitles[$Thumb]}\"" -quality 0 "$WorkingDir/Thumb${Thumb}Label.png"
   done

 # Generate a 'Previous Page' text:

 if [ $CurrentPage -gt 1 ]; then
  convert -size 300x40 -pointsize 24 xc:'#000000FF' -gravity center -fill $FontColor -stroke $FontOutlineColor -strokewidth 3 -draw "font \"$LabelFont\" text 0,0 'Previous Page'" -stroke none -draw "font $LabelFont  text 0,0 'Previous Page'" "$WorkingDir/Previous.png"
 fi

 # Generate a 'Next Page' text:

 if [ $CurrentPage -lt $NumPages ]; then
  convert -size 300x40 -pointsize 24 xc:'#000000FF' -gravity center -fill $FontColor -stroke $FontOutlineColor -strokewidth 3 -draw "font \"$LabelFont\" text 0,0 'Next Page'" -stroke none -draw "font $LabelFont  text 0,0 'Next Page'" "$WorkingDir/Next.png"
 fi

 # Generate Combined Text Layer

 echo "Generating Combination Layer.."

 CompositeCommand="convert \( \("

 for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
  do
  CompositeCommand="$CompositeCommand \( "
  done

 if [ $CurrentPage -gt 1 ]; then
  CompositeCommand="$CompositeCommand \( "
 fi

 if [ $CurrentPage -lt $NumPages ]; then
  CompositeCommand="$CompositeCommand \( "
 fi
 
 CompositeCommand="$CompositeCommand -size ${MenuWidth}x$MenuHeight xc:'#000000FF' "
 
 if [ $CurrentPage -gt 1 ]; then
  CompositeCommand="$CompositeCommand Previous.png -geometry +20+500 -compose over -composite \)"
 fi

 if [ $CurrentPage -lt $NumPages ]; then
  CompositeCommand="$CompositeCommand Next.png -geometry +`echo $[$MenuWidth - 320]`+500 -compose over -composite \) "
 fi

 CompositeCommand="$CompositeCommand Title.png -geometry +0+20 -compose over -composite \) "

 CompositeCommand="$CompositeCommand Subtitle.png -geometry +0+105 -compose over -composite \) "

 for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
  do
  CompositeCommand="$CompositeCommand Thumb${Thumb}Label.png -geometry +$[${ThumbX[$[$Thumb - $FirstScene + 1]]} -  $InterColSpace]+$[${ThumbY[$[$Thumb - $FirstScene + 1]]} + $ThumbHeight + 8 + 15] -compose over -composite \)"
  done

 CompositeCommand="$CompositeCommand -blur 1x0.4 TextLayer.png"

 echo "Executing Composite Command:"
 echo "$CompositeCommand"

 PreviousDir="`pwd`"
 cd "$WorkingDir"
 eval "$CompositeCommand"
 cd "$PreviousDir" 

 if [ $Debug == "false" ]; then

  # Clean Up   

  rm -f "$WorkingDir/Title.png"
  rm -f "$WorkingDir/Subtitle.png"
  rm -f "$WorkingDir/Previous.png"
  rm -f "$WorkingDir/Next.png"

  for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
   do
    rm -f "$WorkingDir/Thumb${Thumb}Label.png"
   done

 fi
}

# ---------------------------------------------------------------
# CREATES ONE FRAME CONTAINING SCENE THUMBNAILS
# ---------------------------------------------------------------

GenerateFrame ()
{

  DiskSpaceWarning "$WorkingDir/Frames/" 10

  # Blend Thumb Frame with Background.

  for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
  do

    # Bounds Check on Blend Factor

    BlendFactor=${BlendFactors[$[$Thumb - $FirstScene + 1]]}
    if [ $BlendFactor -gt 100 ]; then 
     BlendFactor=100
    fi
    if [ $BlendFactor -lt 0 ]; then
     BlendFactor=0
    fi

    FrameIndex=$[$ThumbFrameCounter % ${AvailableFrames[$Thumb]}]
    echo "Using frame $FrameIndex for menu."
    composite -matte -blend $BlendFactor "$WorkingDir/Thumb$Thumb/${ThumbFrameNames[$FrameIndex]}" "$WorkingDir/Thumb${Thumb}Background.png" -quality 0 "$WorkingDir/Thumb${Thumb}Blend.png"
  done

  # Merge the clip frames, labels and title with the backgound:
  # Build one big convert command that merges everything in one go:

  echo
  echo "Merging clip frames with background..."

  CompositeCommand="convert "

 for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
  do
  CompositeCommand="$CompositeCommand \( "
  done

  CompositeCommand="$CompositeCommand BackgroundBlend.jpg "

  for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
    do

    CompositeCommand="$CompositeCommand Thumb${Thumb}Blend.png -geometry +${ThumbX[$[$Thumb - $FirstScene + 1]]}+$[${ThumbY[$[$Thumb - $FirstScene + 1]]} + 15] -compose over -composite \)"

    done

  CompositeCommand="$CompositeCommand -depth 8 +matte Frames/Frame$GeneratedFrameCounter.png"

  #echo "Executing Composite Command: $CompositeCommand"

  PreviousDir="`pwd`"
  cd "$WorkingDir"
  eval "$CompositeCommand"
  cd "$PreviousDir"

  if [ $Debug == "false" ]; then

   # Clean up

   for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
     do
     rm -f "$WorkingDir/Thumb${Thumb}Blend.png"
     done
  fi

  echo -n "Done"
}


# ---------------------------------------------------------------
# UPDATE THE BACKGROUND OF THE MENU 
# ---------------------------------------------------------------

UpdateBackground ()
{
  echo "Updating Background.."

  if [  "$BackgroundType" == "Movie" ]; then

    # Use frame from background movie as background.
    BackgroundThumbFrameNames=( `ls "$WorkingDir/BackgroundMovie/"` )
    if [ ! -z "${BackgroundThumbFrameNames[$[$GeneratedFrameCounter - 1]]}" ]; then
     cp "$WorkingDir/BackgroundMovie/${BackgroundThumbFrameNames[$[$GeneratedFrameCounter - 1]]}" "$WorkingDir/Background.png"
    fi

  else

    if [ "$BackgroundType" == "Checkers" ]; then

      # Create a rolling checkers background

      if [ "WidescreenDVD" == "true" ]; then

        convert -size ${MenuWidth}x`echo $MenuHeight` -roll +`echo $GeneratedFrameCounter`+`echo $GeneratedFrameCounter` tile:"Pictures/lightbluecheckerboardwide.png" -blur 2 "$WorkingDir/Background.png"

      else

        convert -size ${MenuWidth}x`echo $MenuHeight` -roll +`echo $GeneratedFrameCounter`+`echo $GeneratedFrameCounter` tile:"Pictures/lightbluecheckerboard.png" -blur 2 "$WorkingDir/Background.png"

      fi

    else

    rm -f "$WorkingDir/Background.png"
    cp -f "$WorkingDir/backgroundpicture.png" "$WorkingDir/Background.png"

    fi

  fi

  # Create combination layer of Text Layer and Background.

  BlendFactor=${BlendFactors[1]}
  if [ $BlendFactor -gt 100 ]; then 
   BlendFactor=100
  fi
  if [ $BlendFactor -lt 0 ]; then
   BlendFactor=0
  fi

  if [ $WidescreenDVD == "true" ]; then
    WideString="Wide"
  else
    WideString=""
  fi

  # There seems to be a bug in ImageMagick 6.1.x and 6.2.x, the command 
  # "convert test.png -crop 100x100+0+0" cropped.png
  # does not work on png images. Use JPEG in stead.

  convert \( \( "$WorkingDir/Background.png" "Pictures/Shadow$ThumbsPerPage-$(($LastScene - $FirstScene + 1))$WideString.png" -compose over -composite \) "$WorkingDir/TextLayer.png" -compose over -composite \) "$WorkingDir/temp.jpg"

  composite -blend $BlendFactor "$WorkingDir/temp.jpg" "$WorkingDir/Background.png" -depth 8 "$WorkingDir/BackgroundBlend.jpg"

}

# ---------------------------------------------------------------
# CUT OUT THE BACKGROUNDS OF THE THUMBS. NEEDED FOR BLENDING.
# ---------------------------------------------------------------

CreateThumbBackgrounds ()
{
 echo -n "Cutting Background Images"

 for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
 do
   echo -n "."
   convert "$WorkingDir/BackgroundBlend.jpg" -crop ${ThumbWidth}x${ThumbHeight}+${ThumbX[$[$Thumb - $FirstScene + 1]]}+$((${ThumbY[$[$Thumb - $FirstScene + 1]]} + 15)) -quality 0 "$WorkingDir/Thumb${Thumb}Background.png"
 done
 echo
}

# ---------------------------------------------------------------
# FADE BACKGROUND BY BLENDING WITH BLACK FRAME.
# ---------------------------------------------------------------

FadeBackground ()
{
 # Bounds Check
 Fade=$1
 if [ $Fade -gt 100 ]; then 
   Fade=100
 fi

 # size option required???
 eval "composite -blend $Fade \"$WorkingDir/Background.png\" -size ${MenuWidth}x`echo $MenuHeight` xc:black -depth 8 \"$WorkingDir/FadedBackground.png\""
}


# -----------------------------------------------------------------
# GENERATE DVD SUBTITLE BITMAPS FOR MULTIPLEXING WITH MPEG STREAM.
# BITMAPS CONTAIN BORDERS SHOWN FOR CURRENTLY SELECTED SCENE.
# -----------------------------------------------------------------

AddSubtitles ()
{
 # Create empty base images

 if [ $TVSystem == "PAL" ]; then
 convert -size 720x576 xc:'#000000FF' "$WorkingDir/DVDMenuBase.png"
 MenuStrokeWidth=4
 else
 convert -size 720x480 xc:'#000000FF' "$WorkingDir/DVDMenuBase.png"
 MenuStrokeWidth=4
 fi

 convert -size ${MenuWidth}x576 xc:'#000000' "$WorkingDir/DVDMenuSelection.png"

 # Generate image for selected menu items.

 convert "$WorkingDir/DVDMenuSelection.png" -stroke '#FF0000' -fill '#000000' -strokewidth $MenuStrokeWidth -draw "`

  # Generate Image Magick commands to draw a rectangular border around each clip
 for ((Thumb=FirstScene; Thumb <= LastScene ; Thumb++))
 do

  X1=${ThumbX[$[$Thumb - $FirstScene + 1]]}
  Y1=$[${ThumbY[$[$Thumb - $FirstScene + 1]]} + 15]

  X2=$[${ThumbX[$[$Thumb - $FirstScene + 1]]} + $ThumbWidth]
  Y2=$[${ThumbY[$[$Thumb - $FirstScene + 1]]} + $ThumbHeight + 15]

  echo -n "rectangle $X1,$Y1 $X2,$Y2 "

 done 

 # Draw rectangles around 'Previous/Next Page' buttons

 if [ $CurrentPage -gt 1 ]; then

  X1=2
  Y1=500

  X2=300
  Y2=540

  echo -n "rectangle $X1,$Y1 $X2,$Y2 "

 fi

 if [ $CurrentPage -lt $NumPages ]; then

  X1=$(echo $[$MenuWidth - 300])
  Y1=500

  X2=$(echo $[$MenuWidth - 2])
  Y2=540

  echo -n "rectangle $X1,$Y1 $X2,$Y2 "

 fi

 `"  "$WorkingDir/DVDMenuSelection.png"

 if [ $TVSystem == "PAL" ]; then
   convert "$WorkingDir/DVDMenuSelection.png" -filter Point -resize 720x576! "$WorkingDir/DVDMenuSelection.png"
 else
   convert "$WorkingDir/DVDMenuSelection.png" -filter Point -resize 720x480! "$WorkingDir/DVDMenuSelection.png"
 fi

 # Make sure that the images do not contain more than 4 colors

 convert "$WorkingDir/DVDMenuSelection.png" -colors 4 -matte -type PaletteMatte -transparent '#000000'  "$WorkingDir/DVDMenuSelection.png"
 
 cp "$WorkingDir/DVDMenuSelection.png" "$WorkingDir/DVDMenuHighlight.png"

 if [ "$SoundtrackFile" == "" ]; then

   # Generate a bit of silence, the length of the track doesn't matter.

   NumAudioFrames=`echo "scale=2; (48000 / $VideoFramerate) * $TotalFrameCount" | bc | cut -d . -f 1`

   UpdateTask "Generating Empty Soundtrack..."

   pushd "$WorkingDir"

   dd if=/dev/zero bs=2 count=$NumAudioFrames | sox -r 48000 -t .sw - -t .wav -w /dev/stdout | mp2enc -b 128 -r 48000 -o soundtrack.mp2 < /dev/stdin

   popd

 else

   DiskSpaceWarning "$WorkingDir" 100

   PreviousDir="`pwd`"
   cd "$WorkingDir"

   echo "Decoding Soundtrack..."
   UpdateTask "Decoding Soundtrack File..."

   mplayer "$SoundtrackFile" -ao pcm -vc dummy -vo null < /dev/null > /dev/null 2>&1

   cd "$PreviousDir"

   DiskSpaceError "$WorkingDir"

   DetectedSampleRate=`sox -V -t .wav "$WorkingDir/audiodump.wav" -e 2>&1 | grep -E -o [0-9]{1,\}\ samp/sec | awk '{print $1}'`

   Length=`echo "scale=4; tmp = $TotalFrameCount / $VideoFramerate; if (tmp <= 1) tmp = 1; tmp" | bc`
   FadeoutLength=`echo "scale=4; tmp = $Length * 0.05; if (tmp <= 1) tmp=1; tmp" | bc`

   UpdateTask "Processing Soundtrack File..."

   # Trim and Fade-out soundtrack, resample to 48 KHz when needed.

   if [ "$DetectedSampleRate" == 48000 ]; then

     sox "$WorkingDir/audiodump.wav" -t .wav "$WorkingDir/soundtrack.wav" fade t 0 $Length  $FadeoutLength trim 0 $Length

   elif [ -z "$DetectedSampleRate" ]; then

     PrintError "Something is wrong with the soundtrack:\n\n$SoundtrackFile\n\nCould not detect sample rate."

   else

     sox "$WorkingDir/audiodump.wav" -t .wav -r 48000 "$WorkingDir/soundtrack.wav" fade t 0 $Length  $FadeoutLength trim 0 $Length resample -qs 

   fi

   pushd "$WorkingDir"

   cat soundtrack.wav | mp2enc -s -r 48000 -o soundtrack.mp2 < /dev/stdin

   popd

   if [ $Debug == "false" ]; then
    rm -f "$WorkingDir/soundtrack.wav" "$WorkingDir/audiodump.wav"
   fi
 fi

 DiskSpaceError "$WorkingDir"

 echo "Multiplexing.."

 # Write SPUMux Control File

 rm -f "$WorkingDir/spumux.xml"

 echo "<subpictures>" >> "$WorkingDir/spumux.xml"
 echo "<stream>" >> "$WorkingDir/spumux.xml"
 echo "  <spu" >> "$WorkingDir/spumux.xml"
 echo "      force=\"yes\"" >> "$WorkingDir/spumux.xml"
 echo "      start=\"00:00:00.00\"" >> "$WorkingDir/spumux.xml"
 echo "      image=\"$WorkingDir/DVDMenuBase.png\"" >> "$WorkingDir/spumux.xml"
 echo "      select=\"$WorkingDir/DVDMenuSelection.png\"" >> "$WorkingDir/spumux.xml"
 echo "      highlight=\"$WorkingDir/DVDMenuHighlight.png\"" >> "$WorkingDir/spumux.xml"
 echo "      autooutline=\"infer\"" >> "$WorkingDir/spumux.xml"
 echo "      outlinewidth=\"20\"" >> "$WorkingDir/spumux.xml"
 echo "      autoorder=\"rows\">" >> "$WorkingDir/spumux.xml"
 echo "</spu>" >> "$WorkingDir/spumux.xml"
 echo "</stream>" >> "$WorkingDir/spumux.xml"
 echo "</subpictures>" >> "$WorkingDir/spumux.xml"

 # Multiplex Audio, Video and subtitles into one MPEG stream.

 UpdateTask "Multiplexing..."

 mplex -f 8 -o /dev/stdout "$WorkingDir/MenuPage$CurrentPage.m2v" "$WorkingDir/soundtrack.mp2" | spumux "$WorkingDir/spumux.xml" 2>"$WorkingDir/MPlexOutput" > "$WorkingDir/MenuPage$CurrentPage.mpg" 

 echo ""
 echo "MPlex / SPUMux STDERR output:"
 echo ""

 cat "$WorkingDir/MPlexOutput"

 if [ -n "`cat "$WorkingDir/MPlexOutput" | grep "WARN: Bad image"`" ]; then

   PrintError "Your version of ImageMagick fails to cooperate with DVDAuthor.\n\nPlease downgrade to an older version of ImageMagick or upgrade to a newer version of DVDAuthor."

 fi

 if [ -n "`cat "$WorkingDir/MPlexOutput" | grep "ERR:"`" ]; then

   PrintError "MPlex and/or SPUMux generated errors:\n\n`cat "$WorkingDir/MPlexOutput" | grep "ERR:"`"

 fi

}

# ---------------------------------------------------------------
# GENERATE DVD FILE STRUCTURE USING DVDAUTHOR.
# ---------------------------------------------------------------

AuthorDVD ()
{
 UpdateTask "Building DVD file structure..."

 # Clear output directory

 rm -f -R "$WorkingDir/DVD"
 mkdir "$WorkingDir/DVD" 

 # Calculate total diskspace needed to create the DVD file structure

 RequiredDiskSpace=$[ 0`
 for ((Scene=1; Scene <= NumScenes ; Scene++))
  do
  echo -n " + $(du -m "${SceneFiles[$Scene]}" | awk '{print $1}')"
  done

 # Add space needed for menu's

 if [ -f "$WorkingDir/MenuPage1.mpg" ]; then
   for ((CurrentPage=1; CurrentPage <= $NumPages ; CurrentPage++))
     do
       echo -n " + $(du -m "$WorkingDir/MenuPage$CurrentPage.mpg" | awk '{print $1}')"
     done
 fi             ` + 100 ]

 DiskSpaceWarning "$OutputDir" $RequiredDiskSpace


 echo "<?xml version=\"1.0\" encoding=\"iso-8859-1\" ?>" > "$WorkingDir/dvdauthor.xml"
 echo "<dvdauthor dest=\"$WorkingDir/DVD/\">" >> "$WorkingDir/dvdauthor.xml"

 if [ "$IntroMovie" == "" ]; then
   echo " <vmgm />" >> "$WorkingDir/dvdauthor.xml"
 else

   # Play the intro movie right after inserting the DVD

   echo " <vmgm>" >> "$WorkingDir/dvdauthor.xml"
   echo "   <menus>" >> "$WorkingDir/dvdauthor.xml"
   echo "     <pgc>" >> "$WorkingDir/dvdauthor.xml"
   echo "        <vob file=\"$IntroMovie\"/>" >> "$WorkingDir/dvdauthor.xml"
   echo "        <post>" >> "$WorkingDir/dvdauthor.xml"

   # Jump to Menu after playing intro
   # Jump to first title when no menu available.

   if [ -f "$WorkingDir/MenuPage1.mpg" ]; then
   echo "          jump titleset 1 menu;" >> "$WorkingDir/dvdauthor.xml"
   else 
   echo "          jump vmgm title 1;" >> "$WorkingDir/dvdauthor.xml"
   fi

   echo "        </post>" >> "$WorkingDir/dvdauthor.xml"
   echo "     </pgc>" >> "$WorkingDir/dvdauthor.xml"
   echo "   </menus>" >> "$WorkingDir/dvdauthor.xml"
   echo " </vmgm>" >> "$WorkingDir/dvdauthor.xml"

 fi 

 echo " <titleset>" >> "$WorkingDir/dvdauthor.xml"

 # Add the DVD menu pages, when available.

 if [ -f "$WorkingDir/MenuPage1.mpg" ]; then

   echo "File $WorkingDir/MenuPage1.mpg found. DVD will have menus."


   echo "   <menus>" >> "$WorkingDir/dvdauthor.xml"
   echo "     <audio lang=\"en\" />" >> "$WorkingDir/dvdauthor.xml"

   FirstScene=1
   LastScene=$ThumbsPerPage

   # Define all menu pages

   for ((CurrentPage=1; CurrentPage <= $NumPages ; CurrentPage++))
     do
       echo -n "     <pgc" >> "$WorkingDir/dvdauthor.xml"

       # First menu page is the Root menu.

       if [ $CurrentPage -eq 1 ]; then
        echo -n " entry=\"root\"" >> "$WorkingDir/dvdauthor.xml"
       fi

       echo ">" >> "$WorkingDir/dvdauthor.xml"

       echo "       <vob file=\"$WorkingDir/MenuPage$CurrentPage.mpg\" pause=\"2\"></vob>" >> "$WorkingDir/dvdauthor.xml"

       # Add buttons for all thumbnails

       for ((Scene=FirstScene; Scene <= LastScene ; Scene++))
         do
            echo "       <button>jump title $Scene;</button>" >> "$WorkingDir/dvdauthor.xml"
         done

       # Add a PREVIOUS button

       if [ $CurrentPage -gt 1 ]; then
         echo "       <button>jump menu `echo $[$CurrentPage - 1]`;</button>" >> "$WorkingDir/dvdauthor.xml"
       fi

       # Add a NEXT button

       if [ $CurrentPage -lt $NumPages ]; then
         echo "       <button>jump menu `echo $[$CurrentPage + 1]`;</button>" >> "$WorkingDir/dvdauthor.xml"
       fi

       echo "       <post>jump menu ${CurrentPage};</post>" >> "$WorkingDir/dvdauthor.xml"
       echo "     </pgc>" >> "$WorkingDir/dvdauthor.xml"

       FirstScene=$[$FirstScene + $ThumbsPerPage]
       LastScene=$[$LastScene + $ThumbsPerPage]

       if [ $LastScene -gt $NumScenes ]; then
         LastScene=$NumScenes
       fi

     done

   echo "   </menus>" >> "$WorkingDir/dvdauthor.xml"
 fi

 # Add the actual titles

 echo "   <titles>" >> "$WorkingDir/dvdauthor.xml"
 echo "     <audio lang=\"en\" />" >> "$WorkingDir/dvdauthor.xml"
 
 for ((Scene=1; Scene <= NumScenes ; Scene++))
   do
     echo "     <pgc>" >> "$WorkingDir/dvdauthor.xml"
     echo -n "       <vob file=\"${SceneFiles[$Scene]}\" " >> "$WorkingDir/dvdauthor.xml"

     # Add user-configurable pause.
     # Wait 5 seconds after last movie.

     if [ ${ScenePause[$Scene]} != 0 ]; then 
       echo -n "pause=\"${ScenePause[$Scene]}\" " >> "$WorkingDir/dvdauthor.xml"
     elif [ $Scene -eq $NumScenes ]; then
       echo -n "pause=\"5\" " >> "$WorkingDir/dvdauthor.xml"
     fi
     echo  "/>" >> "$WorkingDir/dvdauthor.xml"

     # Jump to next movie, if any

     if [ $Scene -ne $NumScenes ]; then
       echo "       <post> jump title $(($Scene + 1));</post>" >> "$WorkingDir/dvdauthor.xml"
     else 
 
       # Call Root Menu, if we have a menu.

       if [ -f "$WorkingDir/MenuPage1.mpg" ]; then
         echo "       <post> call menu entry root;</post>" >> "$WorkingDir/dvdauthor.xml"
       fi
     fi
     echo "     </pgc>" >> "$WorkingDir/dvdauthor.xml"
   done

 # Call Root Menu, if we have a menu.

 if [ -f "MenuPage1.mpg" ]; then
  echo "       <post>call menu entry root;</post>" >> "$WorkingDir/dvdauthor.xml"
 fi

 echo "   </titles>" >> "$WorkingDir/dvdauthor.xml"
 echo " </titleset>" >> "$WorkingDir/dvdauthor.xml"
 echo "</dvdauthor>" >> "$WorkingDir/dvdauthor.xml"

 dvdauthor -x "$WorkingDir/dvdauthor.xml" > "$WorkingDir/DVDAuthorOutput.txt" 2>&1
 
 echo ""
 echo "DVDAuthor warnings (first 100 lines):"
 echo ""
 echo "------------------------------"
 echo ""

 cat "$WorkingDir/DVDAuthorOutput.txt" | head -n 100

 echo ""
 echo "------------------------------"
 echo ""

 if [ -n "`cat "$WorkingDir/DVDAuthorOutput.txt" | grep ERR:`" ]; then

   PrintError "DVDAuthor failed.\n\nPlease check that your MPEG files are DVD compatible.\n\nThe error message was:\n\n`cat "$WorkingDir/DVDAuthorOutput.txt" | grep ERR:`"

 fi

 DiskSpaceError "$WorkingDir/DVD"

 UpdateTask "DVD is finished."
}


GenerateStaticFrames ()
{
 UpdateTask "Generating static frames..."

 for ((Thumb=1; Thumb <= ThumbsPerPage ; Thumb++))
  do
  BlendFactors[$Thumb]=100
  done

 # Duplicate a frame in the middle of the menu animation
 GeneratedFrameCounter="$[$TotalFrameCount / 2]"

 # Initialise background
 UpdateBackground
 CreateThumbBackgrounds

 # Clean up first
 rm -f -R "$WorkingDir/Frames"
 mkdir "$WorkingDir/Frames"

 # Output Next Frame
 GenerateFrame

 for ((Frame=1; Frame <= $TotalFrameCount ; Frame++))
  do
  ln -s "$WorkingDir/Frames/Frame$GeneratedFrameCounter.png" "$WorkingDir/Frames/Frame$Frame.png"
  done

 UpdateTask "Done."
}

GenerateDynamicFrames ()
{
 # Clean up first
 rm -f -R "$WorkingDir/Frames"
 mkdir "$WorkingDir/Frames"
 
 # Initialise background
 UpdateBackground
 
 #--------------------------------------------------
 # ROUND 1
 # Start by fading in the (animating) background.
 #--------------------------------------------------

 echo -n "Fading in Background"
 UpdateTask "Step 1 of 5: Background appears..."
 for ((Frame=1; Frame <= $NumFadeinFrames ; Frame++,GeneratedFrameCounter++))
  do
   DiskSpaceWarning "$WorkingDir//Frames" 10

   # Make sure the argument runs to 100.
   FadeBackground $[$[$Frame * 150] / $NumFadeinFrames]
 
   mv "$WorkingDir/FadedBackground.png" "$WorkingDir/Frames/Frame$GeneratedFrameCounter.png"

   if [ "$StaticBackground" != "true" ]; then
     UpdateBackground
   fi
   UpdateProgress $[$Frame * 100 / $NumFadeinFrames]
 done

 #--------------------------------------------------
 # ROUND 2 
 # Fade in the animating thumbnails.
 #--------------------------------------------------

 UpdateBackground
 CreateThumbBackgrounds

 # Thumb Frame Counter. 
 ThumbFrameCounter="0"

 UpdateTask "Step 2 of 5: Thumbnails appear..."
 for ((Frame=0; Frame < $ThumbNumFadeinFrames ; Frame++,ThumbFrameCounter++,GeneratedFrameCounter++))
  do

  # Update Background if Necessary
#  if [ "$StaticBackground" != "true" ]; then
   UpdateBackground
   CreateThumbBackgrounds
#  fi

  # Output Next Frame
  GenerateFrame

  # Update Blend Factors
  for ((Thumb=1; Thumb <= ThumbsPerPage ; Thumb++))
   do
   BlendFactors[$Thumb]=$[${BlendFactors[$Thumb]} + $BlendStepSize]
   done

  UpdateProgress $[$Frame * 100 / $ThumbNumFadeinFrames]
 done


 #--------------------------------------------------
 # ROUND 3
 # Just animate the thumbnails and background.
 #--------------------------------------------------

 UpdateBackground

 UpdateTask "Stap 3 of 5: Animating thumbnails ..."
 for ((Frame=0; Frame < $ThumbAnimateFrames ; Frame++,ThumbFrameCounter++,GeneratedFrameCounter++))
  do

  # Update Background if Necessary
  if [ "$StaticBackground" != "true" ]; then
   UpdateBackground
   CreateThumbBackgrounds
  fi

  # Output Next Frame
  GenerateFrame

  UpdateProgress $[$Frame * 100 / $ThumbAnimateFrames]

 done

 #--------------------------------------------------
 # ROUND 4
 # Fade out the clips.
 #--------------------------------------------------

 UpdateTask "Step 4 of 5: Thumbnails disappear..."
 for ((Frame=0; Frame < $ThumbFadeOutFrames ; Frame++,ThumbFrameCounter++,GeneratedFrameCounter++))
  do

  # Update Background if Necessary
#  if [ "$StaticBackground" != "true" ]; then
   UpdateBackground
   CreateThumbBackgrounds
#  fi

  # Output Next Frame
  GenerateFrame

  # Update Blend Factors
  for ((Thumb=1; Thumb <= ThumbsPerPage ; Thumb++))
   do
   BlendFactors[$Thumb]=$[${BlendFactors[$Thumb]} - $BlendStepSize]
   done
  
  UpdateProgress $[$Frame * 100 / $ThumbFadeOutFrames]
  
 done

 #--------------------------------------------------
 # ROUND 5
 # Fade out the Background. End.
 #--------------------------------------------------

 echo "Fading out Background"
 UpdateTask "Step 5 of 5: Background disappears..."
 
 for ((Frame=1; Frame <= $NumFadeinFrames ; Frame++,GeneratedFrameCounter++))
  do

  # Make sure the argument runs to 100.
  FadeBackground $[150 - $[$Frame * 150 / $NumFadeinFrames]]
  mv "$WorkingDir/FadedBackground.png" "$WorkingDir/Frames/Frame$GeneratedFrameCounter.png"

  if [ "$StaticBackground" != "true" ] || [ $Frame == 1 ]; then
   UpdateBackground
  fi
  UpdateProgress $[$Frame * 100 / $NumFadeinFrames]
  done
}



# ---------------------------------------------------------------
# GENERATE ONE MPEG STREAM FOR CURRENT MENU PAGE
# ---------------------------------------------------------------

GeneratePage ()
{
 Initialise

 mkdir -p "$WorkingDir/Frames"
 rm -f "$WorkingDir/Frames/*"

 # Global Frame Counter. Should always start at 1.
 GeneratedFrameCounter="1"

 if [ "$StaticMenu" == "true" ]; then
  GenerateStaticFrames
 else
  GenerateDynamicFrames
 fi
 
 if [ $Debug == "false" ]; then
  
  # Clean up
  
  rm -f "$WorkingDir/Background.png"
  rm -f "$WorkingDir/FadedBackground.png"
 
 
  for ((Thumb=$FirstScene; Thumb <= $LastScene ; Thumb++))
    do
    rm -f "$WorkingDir/Thumb${Thumb}Label.png"
    rm -f "$WorkingDir/Thumb${Thumb}Background.png"
    done

  rm -f "$WorkingDir/TextLayer.png"
 
 fi
 
 UpdateTask "Encoding to DVD Video stream..."
 UpdateProgress 0
 echo "Generating MPEG stream..."

 DiskSpaceWarning "$OutputDir" $[$TotalFrameCount / 20]
 
 PreviousDir="`pwd`"
 cd "$WorkingDir/Frames"
 
 which y4munsharp

 if [ $? -eq 0 ]; then 
 
   # We have MJPEGTools >= 1.8.0
   # Generate CBR stream

   MPEGOptions="--cbr"

 else

   # Generate VBR stream for old MJPEGTools versions

   if [ "$BackgroundType" == "Movie" ]; then
    MPEGOptions=""
   else
    MPEGOptions="-q 6 -H"
   fi

 fi

 if [ $TVSystem == "PAL" ]; then

  TVSystemOption="-n p"

 else

  TVSystemOption="-n n"
  NTSCFramerateOption="-F 4"

 fi

 # Generate a nice MPEG2 stream:

 echo "png2yuv -j Frame%d.png -b 1 -f $VideoFramerate -I p | yuvscaler -O DVD $TVSystemOption | yuvcorrect -Y LUMINANCE_1.0_16_235_16_235 -Y CHROMINANCE_0.0_1.0_128_1.0_128_16_240 | mpeg2enc $NTSCFramerateOption -f 8 -b 7400 -D 10 -M 0 $MPEGOptions $TVSystemOption `

 # Use correct aspect ratio

 if [ $WidescreenDVD == "true" ]; then
  echo "-a 3"
 else
  echo "-a 2"
 fi

 ` -o \"$WorkingDir/MenuPage$CurrentPage.m2v\"" > "$WorkingDir/EncodingScript"

 konsole -e nice bash "$WorkingDir/EncodingScript"

 rm -f "$WorkingDir/EncodingScript"

 DiskSpaceError "$OutputDir"

 cd "$PreviousDir"

 # Add selectable area's to the menu
 AddSubtitles
 UpdateTask "Done."
}


# --------------------
# PROGRAM STARTS HERE
# --------------------

DialogDCOPProgressCommand="$2"
DialogDCOPTaskCommand="$3"

if [ -z "$1" ]; then
 RunChecks
 echo "You did not provide a config file."
 echo
 echo "USAGE: DVDBuilder ConfigFile.conf"
 exit
fi


# ----------------------------------------------------------
# EXTRACT INFO FROM CONFIG FILE
# ----------------------------------------------------------


BackgroundType="$(cat "$1" | grep BackgroundType | cut -d \" -f 2)"
BackgroundMovie="$(cat "$1" | grep BackgroundMovie | cut -d \" -f 2)"
BackgroundPicture="$(cat "$1" | grep BackgroundPicture | cut -d \" -f 2)"
IntroMovie="$(cat "$1" | grep IntroMovie | cut -d \" -f 2)"
SoundtrackFile="$(cat "$1" | grep SoundtrackFile | cut -d \" -f 2)"
NumScenes=$(cat "$1" | grep NumScenes | cut -d = -f 2)
Title="$(cat "$1" | grep Title= | cut -d \" -f 2 | sed "s\'\ \;")"
Subtitle="$(cat "$1" | grep Subtitle= | cut -d \" -f 2 | sed "s\'\ \;")"
TitleFont="$(cat "$1" | grep TitleFont | cut -d \" -f 2)"
SubtitleFont="$(cat "$1" | grep SubtitleFont | cut -d \" -f 2)"
LabelFont="$(cat "$1" | grep LabelFont | cut -d \" -f 2)"
FontColor="$(cat "$1" | grep FontColor | cut -d \" -f 2)"
FontOutlineColor="$(cat "$1" | grep FontOutlineColor | cut -d \" -f 2)"
ThumbDecoration="$(cat "$1" | grep SceneDecoration | cut -d \" -f 2)"
WidescreenDVD="$(cat "$1" | grep Widescreen | cut -d \" -f 2)"
TVSystem="$(cat "$1" | grep TVSystem | cut -d \" -f 2)"
Command="$(cat "$1" | grep Command | cut -d \" -f 2)"
BuildMenu="$(cat "$1" | grep BuildMenu | cut -d \" -f 2)"
StaticMenu="$(cat "$1" | grep StaticMenu | cut -d \" -f 2)"
OutputDir="$(cat "$1" | grep OutputDir | cut -d \" -f 2)"

WorkingDir="$(cat "$1" | grep WorkingDir | cut -d \" -f 2)"

if [ -z "$WorkingDir" ]; then

  PrintError "No working directory given."

fi

if [ "$Command" == "Clean" ]; then

 rm -f -R "$WorkingDir"
 exit 0

fi

mkdir -p "$WorkingDir"


if [ -z "$NumScenes" ] || [ "$NumScenes" -lt 1 ]; then
 PrintError "You did not add any files to the playlist."
fi

for ((Thumb=1; Thumb <= NumScenes ; Thumb++))
 do
 
 SceneFiles[$Thumb]="$( cat "$1" | grep -F SceneFiles[$[$Thumb - 1]]  | cut -d \" -f 2)"
 echo ${SceneFiles[$Thumb]}
 SceneTitles[$Thumb]="$(cat "$1" | grep -F SceneTitles[$[$Thumb - 1]] | cut -d \" -f 2 | sed "s\'\ \;")"

 if [ `echo "${SceneTitles[$Thumb]}" | wc -m` -gt 18 ]; then

   SceneTitles[$Thumb]="`echo "${SceneTitles[$Thumb]}" | head -c 15`..."

 fi

 ScenePause[$Thumb]="$( cat "$1" | grep -F ScenePause[$[$Thumb - 1]]  | cut -d = -f 2)"
 
 done

if [ $BackgroundType == "Picture" ]; then

  if [ -z "$BackgroundPicture" ]; then
    PrintError "You chose to display a picture on the background of the menu, but you did not select one."
  fi

elif [ $BackgroundType == "Movie" ]; then

  if [ -z "$BackgroundMovie" ]; then
    PrintError "You chose to play a soundtrack while the menu is visible, but you did not select one."
  fi

fi


# ----------------------------------------------------------
# SET GLOBAL VARIABLES BASED ON CONFIG READ FROM FILE
# ----------------------------------------------------------

 
if [ $BackgroundType == "Picture" ]; then
  StaticBackground="true"
else
  StaticBackground="false"
fi

# How many clips fit on one menu page?

ThumbsPerPage=$NumScenes
if [ $ThumbsPerPage -gt 6 ]; then
 ThumbsPerPage=6
fi

NumPages=$[$NumScenes / $ThumbsPerPage]
if [ $[$NumScenes % $ThumbsPerPage] != 0 ]; then
 NumPages=$[$NumPages + 1]
fi

echo Outputting $NumPages Pages.


ThumbWidth=150
ThumbHeight=112

# Adjust horizontal Thumb resolution for 16:9 aspect ratio

if [ $WidescreenDVD == "true" ]; then
 ThumbWidth=199
fi


if [ $WidescreenDVD == "true" ]; then
 
 # This odd number is needed for Checkers background.
 # Needs to be divisable by 5

 MenuWidth=1025
else
 MenuWidth=768
fi
MenuHeight=576


if [ $TVSystem == "PAL" ]; then
 VideoFramerate=25
else
 VideoFramerate=29.97
fi

# --------------------------------------------------
# Calculate positions of clips on the screen
# --------------------------------------------------

# Horizontal Space between Thumbnails
InterColSpace=45
# Vertical Space between Rows of Thumbnails
InterRowSpace=55

if [ $ThumbsPerPage == 6 ]; then

ThumbX[1]=`echo "($MenuWidth - 3 * $ThumbWidth - 2 * $InterColSpace) / 2" | bc`
ThumbX[2]=$[${ThumbX[1]} + $ThumbWidth + $InterColSpace]
ThumbX[3]=$[${ThumbX[2]} + $ThumbWidth + $InterColSpace]
ThumbX[4]=${ThumbX[1]}
ThumbX[5]=${ThumbX[2]}
ThumbX[6]=${ThumbX[3]}

ThumbY[1]=`echo "($MenuHeight - 2 * $ThumbHeight - $InterRowSpace ) / 2" | bc`
ThumbY[2]=${ThumbY[1]}
ThumbY[3]=${ThumbY[1]}
ThumbY[4]=$[${ThumbY[1]} + $ThumbHeight + $InterRowSpace]
ThumbY[5]=${ThumbY[4]}
ThumbY[6]=${ThumbY[4]}

BlendFactors[1]=0
BlendFactors[2]=-20
BlendFactors[3]=-40
BlendFactors[4]=-60
BlendFactors[5]=-80
BlendFactors[6]=-100

BlendSteps=200
fi

if [ $ThumbsPerPage == 5 ]; then

ThumbX[1]=`echo "($MenuWidth - 3 * $ThumbWidth - 2 * $InterColSpace) / 2" | bc`
ThumbX[2]=$[${ThumbX[1]} + $ThumbWidth + $InterColSpace]
ThumbX[3]=$[${ThumbX[2]} + $ThumbWidth + $InterColSpace]
ThumbX[4]=`echo "($MenuWidth - 2 * $ThumbWidth - $InterColSpace) / 2" | bc`
ThumbX[5]=$[${ThumbX[4]} + $ThumbWidth + $InterColSpace]

ThumbY[1]=`echo "($MenuHeight - 2 * $ThumbHeight - $InterRowSpace ) / 2" | bc`
ThumbY[2]=${ThumbY[1]}
ThumbY[3]=${ThumbY[1]}
ThumbY[4]=$[${ThumbY[1]} + $ThumbHeight + $InterRowSpace]
ThumbY[5]=${ThumbY[4]}

BlendFactors[1]=0
BlendFactors[2]=-20
BlendFactors[3]=-40
BlendFactors[4]=-60
BlendFactors[5]=-80

BlendSteps=180

fi

if [ $ThumbsPerPage == 4 ]; then

# Create more space between Thumbnails

InterColSpace=75

ThumbX[1]=`echo "($MenuWidth - 2 * $ThumbWidth - $InterColSpace) / 2" | bc`
ThumbX[2]=$[${ThumbX[1]} + $ThumbWidth + $InterColSpace]
ThumbX[3]=${ThumbX[1]}
ThumbX[4]=${ThumbX[2]}

ThumbY[1]=`echo "($MenuHeight - 2 * $ThumbHeight - $InterRowSpace ) / 2" | bc`
ThumbY[2]=${ThumbY[1]}
ThumbY[3]=$[${ThumbY[1]} + $ThumbHeight + $InterRowSpace]
ThumbY[4]=${ThumbY[3]}

BlendFactors[1]=0
BlendFactors[2]=-30
BlendFactors[3]=-60
BlendFactors[4]=-90

BlendSteps=190

fi

if [ $ThumbsPerPage == 3 ]; then

ThumbX[1]=`echo "($MenuWidth - 2 * $ThumbWidth - $InterColSpace) / 2" | bc`
ThumbX[2]=$[${ThumbX[1]} + $ThumbWidth + $InterColSpace]
ThumbX[3]=`echo "($MenuWidth - $ThumbWidth) / 2" | bc`

ThumbY[1]=`echo "($MenuHeight - 2 * $ThumbHeight - $InterRowSpace ) / 2" | bc`
ThumbY[2]=${ThumbY[1]}
ThumbY[3]=$[${ThumbY[1]} + $ThumbHeight + $InterRowSpace]

BlendFactors[1]=0
BlendFactors[2]=-20
BlendFactors[3]=-40

BlendSteps=140

fi

if [ $ThumbsPerPage == 2 ]; then

ThumbX[1]=`echo "($MenuWidth - $ThumbWidth) / 2" | bc`
ThumbX[2]=${ThumbX[1]}

ThumbY[1]=`echo "($MenuHeight - 2 * $ThumbHeight - $InterRowSpace ) / 2" | bc`
ThumbY[2]=$[${ThumbY[1]} + $ThumbHeight + $InterRowSpace]

BlendFactors[1]=0
BlendFactors[2]=-20

BlendSteps=120

fi

if [ $ThumbsPerPage == 1 ]; then

ThumbX[1]=`echo "($MenuWidth - $ThumbWidth) / 2" | bc`
ThumbY[1]=`echo "($MenuHeight - $ThumbHeight) / 2" | bc`

BlendFactors[1]=0
BlendSteps=100

fi

# How many frames of each clip do we use 
# to create the thumbs from?
# This affects the length of the whole menu
# animation. There are 25 frames in one second.

if [ "$Debug" == "true" ]; then
NumSceneFrames=15
else
  if [ $TVSystem == "PAL" ]; then
    NumSceneFrames=400
  else
    NumSceneFrames=480
  fi
fi

# How many frames do we spend fading in the background?
if [ "$Debug" == "true" ]; then
NumFadeinFrames=2
else
  if [ $TVSystem == "PAL" ]; then
    NumFadeinFrames=30
  else
    NumFadeinFrames=36
  fi
fi

# How big a step size to use for fading in thumbs?
if [ "$Debug" == "true" ]; then
BlendStepSize=40
else
  if [ $TVSystem == "PAL" ]; then
    BlendStepSize=6
  else
    BlendStepSize=5
  fi
fi

ThumbNumFadeinFrames=$[$BlendSteps / $BlendStepSize]
ThumbAnimateFrames=$[$NumSceneFrames - $[$ThumbNumFadeinFrames * 2]]
ThumbFadeOutFrames=$[$BlendSteps / $BlendStepSize]

if [ $ThumbsPerPage -le 2 ]; then
  InterColSpace=200
else
  InterColSpace=$[$[${ThumbX[2]}-${ThumbX[1]}-$ThumbWidth] / 2]
fi
TotalFrameCount=$[$[2 * $NumFadeinFrames] + $NumSceneFrames]

RunChecks

if [ $ChecksOK == "false" ]; then
  exit 1
fi

mkdir -p "$WorkingDir"


# Check if all MPEG files are valid

function CheckMPEGFiles

{
 UpdateTask "Checking validity of MPEG files..."
 UpdateProgress 0

 mkdir "$WorkingDir/DVD"

 for ((Thumb=1; Thumb <= NumScenes ; Thumb++))
  do

  dvdauthor -o "$WorkingDir/DVD" -f "${SceneFiles[$Thumb]}" > "$WorkingDir/DVDAuthorOutput.txt" 2>&1

  if [ `cat "$WorkingDir/DVDAuthorOutput.txt" | grep ERR: | wc -l` -gt 0 ]; then

   PrintError "The file\n\n${SceneFiles[$Thumb]}\n\nis not a DVD compliant MPEG2 file."

  elif [ `cat "$WorkingDir/DVDAuthorOutput.txt" | grep WARN: | wc -l` -gt 0 ]; then

   NumWarnings=`cat "$WorkingDir/DVDAuthorOutput.txt" | grep WARN: | wc -l`

   if [ -n "$DISPLAY" ]; then

     kdialog --yesno "The file\n\n${SceneFiles[$Thumb]}\n\ngenerated $NumWarnings warnings. Correct display on DVD players can not be guaranteed.\n\nWould you like to view the warnings?"

     if [ $? == 0 ]; then

       cat "$WorkingDir/DVDAuthorOutput.txt" | grep WARN: > "$WorkingDir/DVDAuthorWarnings.txt"

       kfmclient exec "$WorkingDir/DVDAuthorWarnings.txt"

     fi

   fi

  fi

  UpdateProgress $[ ( 100 * $Thumb ) / $NumScenes ]

  done

 Update Task "All files are OK."
}

if [ "$Command" == "Preview" ]; then

  UpdateProgress 0

  for ((Thumb=1; Thumb <= ThumbsPerPage ; Thumb++))
   do
   BlendFactors[$Thumb]=100
   done

  # Create a frame of page 1, in the middle of the animation
  CurrentPage=1
  ThumbFrameCounter="$[$NumSceneFrames / 2]"
  GeneratedFrameCounter="$[$TotalFrameCount / 2]"
  mkdir -p "$WorkingDir/Frames"

  FirstScene=1
  LastScene=$ThumbsPerPage

  GenerateThumbFrames
  ProcessBackgroundImage

  Initialise
  UpdateProgress 10

  UpdateBackground
  UpdateProgress 20

  CreateThumbBackgrounds
  UpdateProgress 30

  GenerateFrame
  UpdateProgress 70

  mv "$WorkingDir/Frames/Frame$GeneratedFrameCounter.png" "$OutputDir/Preview.png"

  if [ $Debug == "false" ]; then
    rm -f "$WorkingDir/Background.png"
  fi

else

  FirstScene=1
  LastScene=$ThumbsPerPage
  if [ "$StaticMenu" == "true" ]; then
    ThumbFrameCounter="$[$NumSceneFrames / 2]"
  else
    ThumbFrameCounter="0"
  fi

  if [ "$BuildMenu" == "false" ]; then
    echo "We are not building a menu in this run."
  else

    GenerateThumbFrames
    ProcessBackgroundImage

      # Create menu pages

      for ((CurrentPage=1; CurrentPage <= $NumPages ; CurrentPage++))
      do

        GeneratePage

        FirstScene=$[$FirstScene + $ThumbsPerPage]
        LastScene=$[$LastScene + $ThumbsPerPage]

        if [ $LastScene -gt $NumScenes ]; then
          LastScene=$NumScenes
        fi
      done
  fi

  # Build DVD file structure

  if [ "$Command" == "Build" ]; then

    AuthorDVD

  fi

fi

UpdateProgress 100
UpdateTask "Done."

echo ""
echo ""
echo "------------------------"
echo "  END OF SCRIPT RUN"
echo "------------------------"
echo ""
echo ""

# END.